import Immutable from 'immutable';

var getStateName = (function (action) {
  return action && action.type === "@@redux/INIT" ? "initialState argument passed to createStore" : "previous state received by the reducer";
});

var getUnexpectedInvocationParameterMessage = (function (state, reducers, action) {
  var reducerNames = Object.keys(reducers);

  if (!reducerNames.length) {
    return "Store does not have a valid reducer. Make sure the argument passed to combineReducers is an object whose values are reducers.";
  }

  var stateName = getStateName(action);

  if (Immutable.isImmutable ? !Immutable.isImmutable(state) : !Immutable.isCollection(state)) {
    return "The " + stateName + ' is of unexpected type. Expected argument to be an instance of Immutable.Collection or Immutable.Record with the following properties: "' + reducerNames.join('", "') + '".';
  }

  if (Immutable.isMap(state)) {
    var stateMap = state;
    var unexpectedStatePropertyNames = stateMap.filter(function (_, name) {
      return !reducers.hasOwnProperty(name);
    });

    if (unexpectedStatePropertyNames.size > 0) {
      return "Unexpected " + (unexpectedStatePropertyNames.size === 1 ? "property" : "properties") + ' "' + [unexpectedStatePropertyNames.keys()].join('", "') + '" found in ' + stateName + '. Expected to find one of the known reducer property names instead: "' + reducerNames.join('", "') + '". Unexpected properties will be ignored.';
    }
  }

  return null;
});

var combineReducers = (function (reducers, getDefaultState) {
  var reducerKeys = Object.keys(reducers); // eslint-disable-next-line space-infix-ops

  return function (inputState, action) {
    if (typeof inputState === "undefined") {
      inputState = getDefaultState();
    } // eslint-disable-next-line no-process-env


    if (process.env.NODE_ENV !== "production") {
      var warningMessage = getUnexpectedInvocationParameterMessage(inputState, reducers, action);

      if (warningMessage) {
        // eslint-disable-next-line no-console
        console.error(warningMessage);
      }
    }

    return inputState.withMutations(function (temporaryState) {
      reducerKeys.forEach(function (reducerName) {
        var reducer = reducers[reducerName];
        var currentDomainState = temporaryState.get(reducerName);
        var nextDomainState = reducer(currentDomainState, action);

        if (nextDomainState === undefined) {
          throw new Error('Reducer "' + reducerName + '" returned undefined when handling "' + action.type + '" action. To ignore an action, you must explicitly return the previous state.');
        }

        temporaryState.set(reducerName, nextDomainState);
      });
    });
  };
});

export { combineReducers };
